@@ -1,5 +1,5 @@
 /* Fault handler information.  AIX 5 version.
-   Copyright (C) 2005  Bruno Haible <bruno@clisp.org>
+   Copyright (C) 2005, 2009  Bruno Haible <bruno@clisp.org>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,6 +13,6 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
-#include "fault-posix.h"
+#include "fault-posix-ucontext.h"
